"""Helpers."""

import inspect
import logging
from collections.abc import Iterable
from typing import Any, NotRequired, TypedDict

import attrs

_LOGGER = logging.getLogger(__name__)


def as_dict(
    obj: attrs.AttrsInstance, exclude: Iterable[str] | None = None
) -> dict[str, Any]:
    """Represent the object as a dictionary, without empty values and defaults."""

    def _filter(atrb: attrs.Attribute, value: Any) -> bool:
        if exclude and atrb.name in exclude:
            return False
        if atrb.name in ("discovery_extra",):
            return False
        return bool(value) and atrb.default != value and not inspect.isfunction(value)

    res = attrs.asdict(obj, filter=_filter)

    extra = getattr(obj, "discovery_extra", None)
    if extra:
        keys = {k: extra[k] for k in extra if k in res and res[k] != extra[k]}
        _LOGGER.debug("Overwriting %s", keys)
        res.update(extra)

    return res


def hass_abbreviate(
    result: dict[str, Any], *, abbreviations: dict[str, str] | None = None
) -> dict[str, Any]:
    """Abbreviate dictionary keys for entities.

    https://www.home-assistant.io/integrations/mqtt/#supported-abbreviations-in-mqtt-discovery-messages
    """
    if not abbreviations:
        abbreviations = ABBREVIATIONS
    return {abbreviations.get(k, k): v for k, v in result.items()}


def hass_default_rw_icon(*, unit: str) -> str:
    """Get the HASS default icon from the unit."""
    return {
        "W": "mdi:flash",
        "V": "mdi:sine-wave",
        "A": "mdi:current-ac",
        "%": "mdi:battery-lock",
    }.get(unit, "")


def hass_device_class(*, unit: str) -> str:
    """Get the HASS device_class from the unit."""
    return {
        "W": "power",
        "kW": "power",
        "kVA": "apparent_power",
        "VA": "apparent_power",
        "V": "voltage",
        "kWh": "energy",
        "kVAh": "",  # Not energy
        "A": "current",
        "°C": "temperature",
        "%": "battery",
    }.get(unit, "")


class MQTTEntityOptions(TypedDict):
    """Shared MQTTEntity options."""

    name: str
    unique_id: str
    state_topic: str
    object_id: NotRequired[str]

    device_class: NotRequired[str]
    enabled_by_default: NotRequired[bool]
    entity_category: NotRequired[str]
    entity_picture: NotRequired[str]
    expire_after: NotRequired[int]
    icon: NotRequired[str]
    json_attributes_topic: NotRequired[str]
    state_class: NotRequired[str]
    unit_of_measurement: NotRequired[str]

    discovery_extra: NotRequired[dict[str, Any]]


ABBREVIATIONS = {
    v: k
    for k, v in {
        "act_t": "action_topic",
        "act_tpl": "action_template",
        "atype": "automation_type",
        "aux_cmd_t": "aux_command_topic",
        "aux_stat_t": "aux_state_topic",
        "aux_stat_tpl": "aux_state_template",
        "av_tones": "available_tones",
        "avty": "availability",
        "avty_mode": "availability_mode",
        "avty_t": "availability_topic",
        "avty_tpl": "availability_template",
        "away_mode_cmd_t": "away_mode_command_topic",
        "away_mode_stat_t": "away_mode_state_topic",
        "away_mode_stat_tpl": "away_mode_state_template",
        "b_tpl": "blue_template",
        "bri_cmd_t": "brightness_command_topic",
        "bri_cmd_tpl": "brightness_command_template",
        "bri_scl": "brightness_scale",
        "bri_stat_t": "brightness_state_topic",
        "bri_tpl": "brightness_template",
        "bri_val_tpl": "brightness_value_template",
        "clr_temp_cmd_tpl": "color_temp_command_template",
        "clr_temp_cmd_t": "color_temp_command_topic",
        "clr_temp_k": "color_temp_kelvin",
        "clr_temp_stat_t": "color_temp_state_topic",
        "clr_temp_tpl": "color_temp_template",
        "clr_temp_val_tpl": "color_temp_value_template",
        "clrm": "color_mode",
        "clrm_stat_t": "color_mode_state_topic",
        "clrm_val_tpl": "color_mode_value_template",
        "cmd_off_tpl": "command_off_template",
        "cmd_on_tpl": "command_on_template",
        "cmd_t": "command_topic",
        "cmd_tpl": "command_template",
        "cmps": "components",
        "cod_arm_req": "code_arm_required",
        "cod_dis_req": "code_disarm_required",
        "cod_trig_req": "code_trigger_required",
        "cont_type": "content_type",
        "curr_temp_t": "current_temperature_topic",
        "curr_temp_tpl": "current_temperature_template",
        "dev": "device",
        "dev_cla": "device_class",
        "dir_cmd_t": "direction_command_topic",
        "dir_cmd_tpl": "direction_command_template",
        "dir_stat_t": "direction_state_topic",
        "dir_val_tpl": "direction_value_template",
        "dsp_prc": "display_precision",
        "e": "encoding",
        "en": "enabled_by_default",
        "ent_cat": "entity_category",
        "ent_pic": "entity_picture",
        "evt_typ": "event_types",
        "exp_aft": "expire_after",
        "fanspd_lst": "fan_speed_list",
        "flsh": "flash",
        "flsh_tlng": "flash_time_long",
        "flsh_tsht": "flash_time_short",
        "fx_cmd_t": "effect_command_topic",
        "fx_cmd_tpl": "effect_command_template",
        "fx_list": "effect_list",
        "fx_stat_t": "effect_state_topic",
        "fx_tpl": "effect_template",
        "fx_val_tpl": "effect_value_template",
        "fan_mode_cmd_t": "fan_mode_command_topic",
        "fan_mode_cmd_tpl": "fan_mode_command_template",
        "fan_mode_stat_t": "fan_mode_state_topic",
        "fan_mode_stat_tpl": "fan_mode_state_template",
        "frc_upd": "force_update",
        "g_tpl": "green_template",
        "hs_cmd_t": "hs_command_topic",
        "hs_cmd_tpl": "hs_command_template",
        "hs_stat_t": "hs_state_topic",
        "hs_val_tpl": "hs_value_template",
        "ic": "icon",
        "img_e": "image_encoding",
        "img_t": "image_topic",
        "init": "initial",
        "hum_cmd_t": "target_humidity_command_topic",
        "hum_cmd_tpl": "target_humidity_command_template",
        "hum_stat_t": "target_humidity_state_topic",
        "hum_state_tpl": "target_humidity_state_template",
        "json_attr": "json_attributes",
        "json_attr_t": "json_attributes_topic",
        "json_attr_tpl": "json_attributes_template",
        "l_ver_t": "latest_version_topic",
        "l_ver_tpl": "latest_version_template",
        "lrst_t": "last_reset_topic",
        "lrst_val_tpl": "last_reset_value_template",
        "max": "max",
        "max_hum": "max_humidity",
        "max_k": "max_kelvin",
        "max_mirs": "max_mireds",
        "max_temp": "max_temp",
        "migr_discvry": "migrate_discovery",
        "min": "min",
        "min_hum": "min_humidity",
        "min_k": "min_kelvin",
        "min_mirs": "min_mireds",
        "min_temp": "min_temp",
        "mode": "mode",
        "mode_cmd_t": "mode_command_topic",
        "mode_cmd_tpl": "mode_command_template",
        "mode_stat_t": "mode_state_topic",
        "mode_stat_tpl": "mode_state_template",
        "modes": "modes",
        "name": "name",
        "o": "origin",
        "obj_id": "object_id",
        "off_dly": "off_delay",
        "on_cmd_type": "on_command_type",
        "ops": "options",
        "opt": "optimistic",
        "osc_cmd_t": "oscillation_command_topic",
        "osc_cmd_tpl": "oscillation_command_template",
        "osc_stat_t": "oscillation_state_topic",
        "osc_val_tpl": "oscillation_value_template",
        "p": "platform",
        "pct_cmd_t": "percentage_command_topic",
        "pct_cmd_tpl": "percentage_command_template",
        "pct_stat_t": "percentage_state_topic",
        "pct_val_tpl": "percentage_value_template",
        "pl": "payload",
        "pl_arm_away": "payload_arm_away",
        "pl_arm_custom_b": "payload_arm_custom_bypass",
        "pl_arm_home": "payload_arm_home",
        "pl_arm_nite": "payload_arm_night",
        "pl_arm_vacation": "payload_arm_vacation",
        "pl_avail": "payload_available",
        "pl_cln_sp": "payload_clean_spot",
        "pl_cls": "payload_close",
        "pl_dir_fwd": "payload_direction_forward",
        "pl_dir_rev": "payload_direction_reverse",
        "pl_disarm": "payload_disarm",
        "pl_home": "payload_home",
        "pl_inst": "payload_install",
        "pl_loc": "payload_locate",
        "pl_lock": "payload_lock",
        "pl_not_avail": "payload_not_available",
        "pl_not_home": "payload_not_home",
        "pl_off": "payload_off",
        "pl_on": "payload_on",
        "pl_open": "payload_open",
        "pl_osc_off": "payload_oscillation_off",
        "pl_osc_on": "payload_oscillation_on",
        "pl_paus": "payload_pause",
        "pl_stop": "payload_stop",
        "pl_strt": "payload_start",
        "pl_prs": "payload_press",
        "pl_ret": "payload_return_to_base",
        "pl_rst": "payload_reset",
        "pl_rst_hum": "payload_reset_humidity",
        "pl_rst_mode": "payload_reset_mode",
        "pl_rst_pct": "payload_reset_percentage",
        "pl_rst_pr_mode": "payload_reset_preset_mode",
        "pl_stop_tilt": "payload_stop_tilt",
        "pl_stpa": "payload_start_pause",
        "pl_toff": "payload_turn_off",
        "pl_ton": "payload_turn_on",
        "pl_trig": "payload_trigger",
        "pl_unlk": "payload_unlock",
        "pos": "reports_position",
        "pos_clsd": "position_closed",
        "pos_open": "position_open",
        "pr_mode_cmd_t": "preset_mode_command_topic",
        "pr_mode_cmd_tpl": "preset_mode_command_template",
        "pr_mode_stat_t": "preset_mode_state_topic",
        "pr_mode_val_tpl": "preset_mode_value_template",
        "pr_modes": "preset_modes",
        "ptrn": "pattern",
        "r_tpl": "red_template",
        "rel_s": "release_summary",
        "rel_u": "release_url",
        "ret": "retain",
        "rgb_cmd_t": "rgb_command_topic",
        "rgb_cmd_tpl": "rgb_command_template",
        "rgb_stat_t": "rgb_state_topic",
        "rgb_val_tpl": "rgb_value_template",
        "rgbw_cmd_t": "rgbw_command_topic",
        "rgbw_cmd_tpl": "rgbw_command_template",
        "rgbw_stat_t": "rgbw_state_topic",
        "rgbw_val_tpl": "rgbw_value_template",
        "rgbww_cmd_t": "rgbww_command_topic",
        "rgbww_cmd_tpl": "rgbww_command_template",
        "rgbww_stat_t": "rgbww_state_topic",
        "rgbww_val_tpl": "rgbww_value_template",
        "send_cmd_t": "send_command_topic",
        "send_if_off": "send_if_off",
        "set_fan_spd_t": "set_fan_speed_topic",
        "set_pos_t": "set_position_topic",
        "set_pos_tpl": "set_position_template",
        "pos_t": "position_topic",
        "pos_tpl": "position_template",
        "spd_rng_min": "speed_range_min",
        "spd_rng_max": "speed_range_max",
        "src_type": "source_type",
        "stat_cla": "state_class",
        "stat_closing": "state_closing",
        "stat_clsd": "state_closed",
        "stat_jam": "state_jammed",
        "stat_locked": "state_locked",
        "stat_locking": "state_locking",
        "stat_off": "state_off",
        "stat_on": "state_on",
        "stat_open": "state_open",
        "stat_opening": "state_opening",
        "stat_stopped": "state_stopped",
        "stat_unlocked": "state_unlocked",
        "stat_unlocking": "state_unlocking",
        "stat_t": "state_topic",
        "stat_tpl": "state_template",
        "stat_val_tpl": "state_value_template",
        "step": "step",
        "stype": "subtype",
        "sug_dsp_prc": "suggested_display_precision",
        "sup_clrm": "supported_color_modes",
        "sup_dur": "support_duration",
        "sup_vol": "support_volume_set",
        "sup_feat": "supported_features",
        "swing_mode_cmd_t": "swing_mode_command_topic",
        "swing_mode_cmd_tpl": "swing_mode_command_template",
        "swing_mode_stat_t": "swing_mode_state_topic",
        "swing_mode_stat_tpl": "swing_mode_state_template",
        "t": "topic",
        "temp_cmd_t": "temperature_command_topic",
        "temp_cmd_tpl": "temperature_command_template",
        "temp_hi_cmd_t": "temperature_high_command_topic",
        "temp_hi_cmd_tpl": "temperature_high_command_template",
        "temp_hi_stat_t": "temperature_high_state_topic",
        "temp_hi_stat_tpl": "temperature_high_state_template",
        "temp_lo_cmd_t": "temperature_low_command_topic",
        "temp_lo_cmd_tpl": "temperature_low_command_template",
        "temp_lo_stat_t": "temperature_low_state_topic",
        "temp_lo_stat_tpl": "temperature_low_state_template",
        "temp_stat_t": "temperature_state_topic",
        "temp_stat_tpl": "temperature_state_template",
        "temp_unit": "temperature_unit",
        "tilt_clsd_val": "tilt_closed_value",
        "tilt_cmd_t": "tilt_command_topic",
        "tilt_cmd_tpl": "tilt_command_template",
        "tilt_max": "tilt_max",
        "tilt_min": "tilt_min",
        "tilt_opnd_val": "tilt_opened_value",
        "tilt_opt": "tilt_optimistic",
        "tilt_status_t": "tilt_status_topic",
        "tilt_status_tpl": "tilt_status_template",
        "tit": "title",
        "trns": "transition",
        "uniq_id": "unique_id",
        "unit_of_meas": "unit_of_measurement",
        "url_t": "url_topic",
        "url_tpl": "url_template",
        "val_tpl": "value_template",
        "whit_cmd_t": "white_command_topic",
        "whit_scl": "white_scale",
        "xy_cmd_t": "xy_command_topic",
        "xy_cmd_tpl": "xy_command_template",
        "xy_stat_t": "xy_state_topic",
        "xy_val_tpl": "xy_value_template",
    }.items()
}



"""Example Options for HASS addon. Extend Options."""

import logging
import os
import typing as t
from json import load, loads
from json.decoder import JSONDecodeError
from pathlib import Path

import attrs
from cattrs import Converter, transform_error
from cattrs.gen import make_dict_structure_fn
from yaml import safe_load

from . import supervisor
from .utils import logging_color

_LOGGER = logging.getLogger(__name__)


@attrs.define()
class AddonOptions:
    """HASS Addon Options."""

    def load_dict(
        self, value: dict, log_lvl: int = logging.DEBUG, log_msg: str = ""
    ) -> None:
        """Structure and copy result to self."""
        try:
            _LOGGER.log(log_lvl, "%s: %s", log_msg or "Loading config", value)
            val = CONVERTER.structure(value, self.__class__)
        except Exception as exc:
            msg = "Error loading config: " + "\n".join(transform_error(exc))
            _LOGGER.error(msg)
            raise ValueError(msg) from None
        for key in value:
            setattr(self, key.lower(), getattr(val, key.lower()))

    def load_env(self) -> bool:
        """Get attrs fields from the environment."""
        res = {}
        atts: tuple[attrs.Attribute, ...] = attrs.fields(
            attrs.resolve_types(self.__class__)
        )
        for att in atts:
            val = os.getenv(att.name.upper())
            if not val:
                continue
            name = att.name.lower()
            if att.type is list or t.get_origin(att.type) is list:
                try:
                    res[name] = loads(val)
                except JSONDecodeError:
                    res[name] = val.split(",")
                continue
            res[name] = val
        if res:
            self.load_dict(
                res,
                logging.INFO,
                "Loading config from environment variables",
            )
        return bool(res)

    async def init_addon(self) -> None:
        """Initialize options from environment variables and config files & setup the logger."""
        logging_color()

        env_ok = self.load_env()

        cfg_names = (
            "/data/options.json",  # HA OS config location
            "/data/options.yaml",  # Recommended for standalone deployment
            ".data/options.yaml",  # Pytest
        )
        cfg_files = [f for f in (Path(s) for s in cfg_names) if f.exists()]
        if not cfg_files and not env_ok:
            _LOGGER.error("No config file or environment variables found.")
            os._exit(1)

        for fpath in cfg_files:
            _LOGGER.info("Loading config: %s", fpath)
            with fpath.open("r", encoding="utf-8") as fptr:
                opt = load(fptr) if fpath.suffix == ".json" else safe_load(fptr)
                self.load_dict(opt)

        if getattr(self, "debug", 0):
            logging_color(debug=True)


MQFAIL = "MQTT: Failed to get MQTT service details from the Supervisor"


@attrs.define()
class MQTTOptions(AddonOptions):
    """MQTT Options for HASS Addon."""

    mqtt_host: str = "core-mosquitto"
    mqtt_port: int = 1883
    mqtt_username: str = ""
    mqtt_password: str = ""

    async def init_addon(self) -> None:
        """Initialize MQTT options from environment variables and config files."""
        await super().init_addon()

        # Don't warn if MQTT password is set
        if not supervisor.token(warn=False):
            _LOGGER.log(
                logging.DEBUG if self.mqtt_password else logging.WARNING, MQFAIL
            )
            return

        data = await supervisor.get("/services/mqtt")
        if not data:
            return

        try:
            data = t.cast(dict[str, t.Any], data["data"])
            self.mqtt_host = data["host"]
            self.mqtt_port = data["port"]
            self.mqtt_username = data["username"]
            self.mqtt_password = data["password"]
        except KeyError as err:
            _LOGGER.warning("%s: %s %s", MQFAIL, err, data)


CONVERTER = Converter(forbid_extra_keys=True)


def structure_ensure_lowercase_keys(cls: type) -> t.Callable[[t.Any, t.Any], t.Any]:
    """Convert any uppercase keys to lowercase."""
    struct = make_dict_structure_fn(cls, CONVERTER)  # type: ignore[var-annotated]

    def structure(d: dict[str, t.Any], cl: t.Any) -> t.Any:
        lower = [k for k in d if k.lower() != k]
        for k in lower:
            if k.lower() in d:
                _LOGGER.warning("Key %s already exists in lowercase", k.lower())
            d[k.lower()] = d.pop(k)
        return struct(d, cl)

    return structure


CONVERTER.register_structure_hook_factory(attrs.has, structure_ensure_lowercase_keys)



"""Home Assistant Supervisor helpers."""

import logging
import os
from typing import Any
from urllib.parse import urljoin

from aiohttp import ClientSession

_LOGGER = logging.getLogger(__name__)

MQFAIL = "Supervisor: Failed to get service details from the Supervisor"


def token(warn: bool = True, fail: bool = False) -> str | None:
    """Get the SUPERVISOR_TOKEN.

    Requires:
        hassio_api: true
        services: ["mqtt:need"]
    """
    token = os.environ.get("SUPERVISOR_TOKEN")
    if not token and warn:
        _LOGGER.error("Supervisor: No SUPERVISOR_TOKEN. Check addon config.")
        if not fail:
            return None
        raise ValueError("Supervisor: No SUPERVISOR_TOKEN. Check addon config.")
    return token


async def get(url: str) -> dict[str, Any] | None:
    """Get json data from the HA Supervisor."""
    url = urljoin("http://supervisor", url)
    head = {
        "Authorization": f"Bearer {token(fail=True)}",
        "content-type": "application/json",
    }
    async with ClientSession() as session:
        async with session.get(url, headers=head) as res:
            if res.status != 200:
                _LOGGER.warning("Supervisor: get %s, response %s", url, res.status)
                return None
            return await res.json()



"""MQTTClient."""

import asyncio
import importlib.metadata
import inspect
import logging
import time
from collections.abc import Callable, Coroutine
from json import dumps
from typing import Any

import attrs
from paho.mqtt.client import Client, MQTTMessage
from paho.mqtt.enums import CallbackAPIVersion
from paho.mqtt.reasoncodes import ReasonCode

from .device import MQTTDevice, MQTTOrigin
from .utils import load_json

HA_STATUS_TOPIC = "homeassistant/status"
_LOGGER = logging.getLogger(__name__)
MQTT_EXPLORER_LIMIT = 20000

type TopicCallback = (
    Callable[[str], None]
    | Callable[[str, str], None]
    | Callable[[str], Coroutine[Any, Any, None]]
    | Callable[[str, str], Coroutine[Any, Any, None]]
)


@attrs.define()
class MQTTAsyncClient:
    """Async MQTT Client."""

    availability_topic: str = ""
    client: Client = attrs.field(init=False, repr=False)
    topics_subscribed: dict[str, Callable] = attrs.field(init=False, repr=False)
    """All topics subscribed to."""

    connect_time: float = attrs.field(init=False, repr=False)

    def __attrs_post_init__(self) -> None:
        """Init."""
        self.topics_subscribed = {}
        self.connect_time = 0
        self.client = Client(callback_api_version=CallbackAPIVersion.VERSION2)
        self.client.on_connect = self._mqtt_on_connect
        self.client.on_message = self._mqtt_on_message

    async def connect(
        self,
        options: Any = None,
        *,
        username: str = "",
        password: str = "",
        host: str = "core-mosquitto",
        port: int = 1883,
        wait_connected: bool = False,
    ) -> None:
        """Connect to MQTT server specified as attributes of the options."""
        if self.client.is_connected():
            _LOGGER.warning("MQTT: Client connected. Reconnecting...")
        await self.disconnect()  # "Connection Successful" triggered on re-connect

        if options:
            username = getattr(options, "mqtt_username", username)
            password = getattr(options, "mqtt_password", password)
            host = getattr(options, "mqtt_host", host)
            port = getattr(options, "mqtt_port", port)
        self.client.username_pw_set(username=username, password=password)

        if self.availability_topic:
            self.client.will_set(self.availability_topic, "offline", retain=True)

        _LOGGER.info("MQTT: Connecting to %s@%s:%s", username, host, port)
        self.client.connect_async(host=host, port=port)
        self.client.loop_start()
        self.connect_time = time.time() + 5

        if wait_connected:
            await self.wait_connected()

    def _mqtt_on_connect(
        self,
        client: Client,
        data: Any,
        flags: Any,
        rc: ReasonCode,
        prop: Any = None,
    ) -> None:
        """MQTT on_connect callback."""
        if rc != 0:
            _LOGGER.error("MQTT: Connection failed with reason code %s", rc)
            self.connect_time = -1  # failed
            return
        _LOGGER.info("MQTT: Connected")
        # publish online (Last will sets offline on disconnect)
        if self.availability_topic:
            client.publish(self.availability_topic, "online", retain=True)
        # Subscribe to all existing change handlers (on connect/reconnect)
        for topic in self.topics_subscribed:
            client.subscribe(topic)

    async def wait_connected(self) -> None:
        """Wait until connected."""
        if self.client.is_connected():
            return
        if self.connect_time == 0:
            raise RuntimeError("MQTT: Call connect first")
        _LOGGER.debug("MQTT: Waiting for connection...")
        while True:
            if self.client.is_connected():
                return
            await asyncio.sleep(0.1)
            if time.time() > self.connect_time:
                if self.connect_time < 0:
                    raise ConnectionError("MQTT: Connection failed")
                msg = "MQTT: Connection timeout (5s)"
                _LOGGER.error(msg)
                raise ConnectionError(msg)

    async def disconnect(self) -> None:
        """Stop the MQTT client."""

        def _stop() -> None:
            """Do not disconnect, allow the broker to publish LWT message."""
            self.client.loop_stop()

        await asyncio.get_running_loop().run_in_executor(None, _stop)

    def publish_args(
        self, topic: str, payload: str | None, qos: int, retain: bool
    ) -> tuple[str, str | None, int, bool]:
        """Prep publish parameters."""
        if not topic:
            raise ValueError(f"MQTT: Cannot publish to empty topic (payload={payload})")
        if not isinstance(qos, int):
            qos = 0
        if retain:
            qos = 1
        _LOGGER.debug(
            "MQTT: Publish %s%s %s, %s", qos, "R" if retain else "", topic, payload
        )
        if payload and len(payload) > MQTT_EXPLORER_LIMIT:
            _LOGGER.info(
                "MQTT: Payload >%s: %s (MQTTExplorer will truncate the message)",
                MQTT_EXPLORER_LIMIT,
                len(payload),
            )
        return (topic, payload, qos, bool(retain))

    async def publish(
        self,
        topic: str,
        payload: str | None = None,
        qos: int = 0,
        retain: bool = False,
    ) -> None:
        """Publish a MQTT message."""
        args = self.publish_args(topic, payload, qos, retain)
        await self.wait_connected()
        await asyncio.get_running_loop().run_in_executor(
            None, self.client.publish, *args
        )

    def topic_unsubscribe(self, topic: str) -> None:
        """Remove a topic from the topic callbacks."""
        self.client.unsubscribe(topic)
        self.client.message_callback_remove(topic)
        self.topics_subscribed.pop(topic, None)

    def topic_subscribe(self, topic: str, callback: TopicCallback) -> None:
        """Add a topic to the topic callbacks."""
        if topic in self.topics_subscribed:
            _LOGGER.warning("MQTT: Topic %s already subscribed", topic)
            return

        paramc = len(inspect.signature(callback).parameters)
        loop = asyncio.get_running_loop()

        def cb(client: Client, _userdata: Any, message: MQTTMessage) -> None:
            """Topic's callback."""
            payload = message.payload.decode("utf-8")
            _LOGGER.debug(
                "MQTT: Callback for topic %s, payload %s",
                message.topic,
                payload or '""',
            )
            args = [] if paramc == 1 else [message.topic]
            if inspect.iscoroutinefunction(callback):
                coro = callback(payload, *args)
                loop.call_soon_threadsafe(lambda: loop.create_task(coro))
            else:
                callback(payload, *args)

        _LOGGER.debug("MQTT: Add callback for topic %s", topic)
        self.client.subscribe(topic)
        self.client.message_callback_add(topic, cb)
        self.topics_subscribed[topic] = cb

    def _mqtt_on_message(self, c: Client, userdata: Any, message: MQTTMessage) -> None:
        """MQTT on_message fallback."""
        t = message.topic
        p = message.payload.decode("utf-8")
        _LOGGER.warning("MQTT: Unhandled msg received. Topic %s with payload %s", t, p)


@attrs.define()
class MQTTClient(MQTTAsyncClient):
    """Home Assistant specific MQTT client."""

    devs: list[MQTTDevice] = attrs.field(factory=list)

    origin_name: str = "mqtt-entity"
    origin_version: str = attrs.field(
        factory=lambda: importlib.metadata.version("mqtt-entity")
    )
    origin_url: str = ""
    clean_entities: int = attrs.field(default=1)
    """Clean entities on discovery: 1=migrate, 2=remove, 0=none."""

    def monitor_homeassistant_status(self) -> None:
        """Monitor homeassistant/status & publish discovery info."""
        if HA_STATUS_TOPIC in self.topics_subscribed:
            return
        _loop = asyncio.get_running_loop()

        def _timeout() -> None:
            _LOGGER.warning(
                "MQTT: Timeout waiting for Home Assistant. The %s topic is empty.\n"
                "Configure the MQTT integration in Home Assistant to publish a "
                "last will & testament (online/offline) with the Retain flag set.",
                HA_STATUS_TOPIC,
            )
            _LOGGER.warning(
                "MQTT: Your entities will be unavailable if HA restarts",
            )
            _loop.create_task(self.publish_discovery_info())  # noqa: RUF006

        timeout = _loop.call_later(10, _timeout)

        async def _online_cb(payload_s: str) -> None:
            """Republish discovery info."""
            if payload_s != "online":
                _LOGGER.warning(
                    "MQTT: Home Assistant offline. %s = %s", HA_STATUS_TOPIC, payload_s
                )
                return
            timeout.cancel()
            _LOGGER.info(
                "MQTT: Home Assistant online. Publish discovery info for %s",
                [d.name for d in self.devs],
            )
            await self.publish_discovery_info()

        self.topic_subscribe(HA_STATUS_TOPIC, _online_cb)
        if self.connect_time == 0:
            raise ConnectionError()

    async def publish_discovery_info(self) -> None:
        """Publish discovery info immediately."""
        if not self.devs:
            _LOGGER.warning("MQTT: No devices to publish discovery info for")
            return

        if self.clean_entities:
            self._clean_entity_based_discovery()
            await asyncio.sleep(1)

        for ddev in self.devs:
            disco_topic, disco_dict = ddev.discovery_info(
                self.availability_topic,
                origin=MQTTOrigin(
                    name=self.origin_name,
                    sw=self.origin_version,
                    url=self.origin_url,
                ),
            )
            disco_payload = dumps(disco_dict)
            if len(disco_payload) > MQTT_EXPLORER_LIMIT:
                disco_payload = dumps(disco_dict, indent=None, separators=(",", ":"))
            await self.publish(disco_topic, disco_payload)

            # add topic callbacks
            tcb: dict[str, TopicCallback] = {}
            for ent in ddev.components.values():
                tcb.update(ent.topic_callbacks)
            for topic, cbk in tcb.items():
                self.topic_subscribe(topic, cbk)

    def _clean_entity_based_discovery(self) -> None:
        """Remove entity based discovery as part of discovery info.

        https://www.home-assistant.io/docs/mqtt/discovery/
        Publish discovery topics on "homeassistant/device/{device_id}/{sensor_id}/config"
        Publish discovery topics on "homeassistant/(sensor|switch)/{device_id}/{sensor_id}/config"
        """

        async def cb_migrate(payload_s: str, topic: str) -> None:
            """Migrate to device based discovery."""
            if not payload_s:
                return
            payload = load_json(payload_s)
            _LOGGER.info("MQTT MIGRATE topic %s with payload %s", topic, payload)
            migrate_ok = payload == {"migrate_discovery": True}
            _pl = None if migrate_ok else dumps({"migrate_discovery": True})
            if migrate_ok:
                await asyncio.sleep(5)
            await self.publish(topic=topic, payload=_pl, qos=1, retain=True)

        def cb_remove(dev: MQTTDevice) -> TopicCallback:
            """Create a callback for the device."""

            async def _cb_remove(payload_s: str, topic: str) -> None:
                if not payload_s:
                    return
                payload = load_json(payload_s)
                # if not part of this device, remove the topic
                if not isinstance(payload, dict) or "unique_id" not in payload:
                    _LOGGER.warning(
                        "MQTT CLEAN: No unique_id in payload %s, cannot remove", payload
                    )
                    return
                uid = payload["unique_id"]
                if uid not in dev.components:
                    _LOGGER.info("MQTT: Removing unique ID %s", uid)
                    self.client.publish(topic=topic, payload=None, qos=1, retain=True)

            return _cb_remove

        if self.clean_entities == 0:
            return
        migrate = self.clean_entities == 1
        self.clean_entities = 0
        for dev in self.devs:
            topic = f"homeassistant/+/{dev.id}/+/config"
            self.topic_subscribe(topic, cb_migrate if migrate else cb_remove(dev))
            asyncio.get_running_loop().call_later(10, self.topic_unsubscribe, topic)



"""HASS MQTT Device, used for device based discovery."""

from typing import Any

import attrs
from attrs import validators

from .entities import MQTTBaseEntity
from .helpers import as_dict, hass_abbreviate


@attrs.define()
class MQTTOrigin:
    """Represent the origin of an MQTT message."""

    name: str
    sw: str = ""
    """ws_version"""
    url: str = ""
    """support_url"""


@attrs.define()
class MQTTDevice:
    """Base class for MQTT Device Discovery. A Home Assistant Device groups entities."""

    identifiers: list[str | tuple[str, Any]] = attrs.field(
        validator=[validators.instance_of(list), validators.min_len(1)]
    )

    components: dict[str, MQTTBaseEntity]
    """MQTT component entities."""
    remove_components: dict[str, str] = attrs.field(factory=dict)
    """Components to be removed on discovery. object_id and the platform name."""

    connections: list[str] = attrs.field(factory=list)
    configuration_url: str = ""
    manufacturer: str = ""
    model: str = ""
    name: str = ""
    suggested_area: str = ""
    sw_version: str = ""
    via_device: str = ""

    @property
    def id(self) -> str:
        """The device identifier. Also object_id."""
        return str(self.identifiers[0])

    def discovery_info(
        self, availability_topic: str, *, origin: MQTTOrigin
    ) -> tuple[str, dict[str, Any]]:
        """Return the discovery dictionary for the MQTT device."""
        cmps = {
            k: hass_abbreviate(v.as_discovery_dict) for k, v in self.components.items()
        }
        for key, platform in self.remove_components.items():
            cmps[key] = {"p": cmps[key]["p"] if key in cmps else platform}

        return (
            f"homeassistant/device/{self.id}/config",
            {
                "dev": as_dict(self, exclude=["components", "remove_components"]),
                "o": as_dict(origin),
                "avty": {"topic": availability_topic},
                "cmps": cmps,
            },
        )



"""mqtt-entity library."""

from mqtt_entity.client import MQTTClient
from mqtt_entity.device import MQTTDevice
from mqtt_entity.entities import (
    MQTTBaseEntity,
    MQTTBinarySensorEntity,
    MQTTDeviceTrigger,
    MQTTEntity,
    MQTTLightEntity,
    MQTTNumberEntity,
    MQTTRWEntity,
    MQTTSelectEntity,
    MQTTSensorEntity,
    MQTTSwitchEntity,
    MQTTTextEntity,
)

__all__ = [
    "MQTTBaseEntity",
    "MQTTBinarySensorEntity",
    "MQTTClient",
    "MQTTDevice",
    "MQTTDeviceTrigger",
    "MQTTEntity",
    "MQTTLightEntity",
    "MQTTNumberEntity",
    "MQTTRWEntity",
    "MQTTSelectEntity",
    "MQTTSensorEntity",
    "MQTTSwitchEntity",
    "MQTTTextEntity",
]



"""Utilities."""

import logging
from json import loads
from json.decoder import JSONDecodeError
from math import modf
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import attrs

BOOL_ON = "ON"
BOOL_OFF = "OFF"


def load_json(msg: str | None) -> dict[str, Any] | str:
    """Load a JSON string into a dictionary."""
    if not msg:
        return {}
    try:
        res = loads(msg)
        if isinstance(res, dict):
            return res
    except JSONDecodeError:
        pass
    return str(msg)


def logging_color(*, debug: bool = False, force: bool = True) -> None:
    """Enable color logging."""
    try:
        import colorlog  # noqa: PLC0415

        handler = colorlog.StreamHandler()
        handler.setFormatter(
            colorlog.ColoredFormatter(
                "[%(asctime)s] %(log_color)s%(levelname)-7s%(reset)s %(message)s",
                datefmt="%H:%M:%S",
                reset=False,
            )
        )
        logging.basicConfig(
            level=logging.DEBUG if debug else logging.INFO,
            handlers=[handler],
            force=force,
        )
    except ModuleNotFoundError:
        logging.basicConfig(
            format="%(asctime)s %(levelname)-7s %(name)s %(message)s",
            level=logging.DEBUG if debug else logging.INFO,
            force=force,
        )


def required(_obj: Any, attr_obj: "attrs.Attribute[Any]", val: Any) -> None:
    """Ensure an attrs.field is present."""
    if val is None:
        raise TypeError(f"Argument '{getattr(attr_obj, 'name', '')}' missing")


def tostr(val: Any) -> str:
    """Convert a value to a string with maximum 3 decimal places."""
    if isinstance(val, str):
        return val
    if val is None:
        return ""
    if isinstance(val, bool):
        return BOOL_ON if val else BOOL_OFF
    if not isinstance(val, float):
        return str(val)
    if modf(val)[0] == 0:
        return str(int(val))
    return f"{val:.3f}".rstrip("0")



"""MQTT entities."""

from __future__ import annotations

from collections.abc import Callable, Sequence
from json import dumps
from typing import TYPE_CHECKING, Any

import attrs
from attrs import validators

from .helpers import as_dict
from .utils import BOOL_OFF, BOOL_ON, required, tostr

if TYPE_CHECKING:
    from .client import MQTTAsyncClient, TopicCallback


@attrs.define()
class MQTTBaseEntity:
    """Base class for entities that support MQTT Discovery."""

    @property
    def topic_callbacks(self) -> dict[str, TopicCallback]:
        """Append topics and callbacks."""
        return {}

    @property
    def as_discovery_dict(self) -> dict[str, Any]:
        """Discovery dict."""
        return as_dict(self)


@attrs.define()
class MQTTEntity(MQTTBaseEntity):
    """A generic Home Assistant entity used as the base class for other entities."""

    name: str
    unique_id: str
    """An ID that uniquely identifies this sensor. If two sensors have the same unique ID,
      Home Assistant will raise an exception. Required when used with device-based discovery."""
    state_topic: str
    object_id: str = ""
    """Used instead of name for automatic generation of entity_id."""

    device_class: str = ""
    """The type/class of the sensor to set the icon in the frontend. The device_class can be null."""
    enabled_by_default: bool = True
    """Flag which defines if the entity should be enabled when first added."""
    entity_category: str = ""
    """The category of the entity. When set, the entity category must be diagnostic for sensors."""
    entity_picture: str = ""
    """Picture URL for the entity."""
    expire_after: int = 0
    """If set, it defines the number of seconds after the sensor's state expires,
    if it's not updated. After expiry, the sensor's state becomes unavailable.
    Default the sensors state never expires."""
    icon: str = ""
    json_attributes_topic: str = ""
    state_class: str = ""
    """https://developers.home-assistant.io/docs/core/entity/sensor/#available-state-classes"""
    unit_of_measurement: str = ""
    """Defines the units of measurement of the sensor, if any. The unit_of_measurement can be null."""

    discovery_extra: dict[str, Any] = attrs.field(factory=dict)
    """Additional MQTT Discovery attributes."""

    platform = ""

    async def send_state(
        self, client: MQTTAsyncClient, payload: Any, *, retain: bool = False
    ) -> None:
        """Publish the state to the MQTT state topic."""
        await client.publish(self.state_topic, tostr(payload), retain=retain)

    async def send_json_attributes(
        self,
        client: MQTTAsyncClient,
        attributes: dict[str, Any],
        *,
        retain: bool = True,
    ) -> None:
        """Publish the attributes to the MQTT JSON attributes topic."""
        await client.publish(
            topic=self.json_attributes_topic, payload=dumps(attributes), retain=retain
        )

    def __attrs_post_init__(self) -> None:
        """Init the class."""
        if not self.platform:
            raise TypeError(f"Do not instantiate {self.__class__.__name__} directly")
        if not self.state_class and self.device_class == "energy":
            self.state_class = "total_increasing"

    @property
    def as_discovery_dict(self) -> dict[str, Any]:
        """Discovery dict."""
        res = super().as_discovery_dict
        res.setdefault("platform", self.platform)
        return res


@attrs.define()
class MQTTDeviceTrigger(MQTTBaseEntity):
    """A Home Assistant Device trigger.

    https://www.home-assistant.io/integrations/device_trigger.mqtt/
    """

    type: str
    subtype: str
    payload: str
    topic: str

    @property
    def name(self) -> str:
        """Return the name of the trigger."""
        return f"{self.type} {self.subtype}".strip()

    async def send_trigger(self, client: MQTTAsyncClient) -> None:
        """Publish the state to the MQTT state topic."""
        await client.publish(self.topic, self.payload or "1")

    discovery_extra: dict[str, Any] = attrs.field(factory=dict)
    """Additional MQTT Discovery attributes."""

    @property
    def as_discovery_dict(self) -> dict[str, Any]:
        """Return the final discovery dictionary."""
        result = super().as_discovery_dict
        result["automation_type"] = "trigger"
        result["platform"] = "device_automation"
        return result


@attrs.define()
class MQTTRWEntity(MQTTEntity):
    """Read/Write entity base class.

    This will default to a text entity.
    """

    command_topic: str = attrs.field(
        default="", validator=(validators.instance_of(str), validators.min_len(2))
    )
    on_command: Callable | None = None

    platform = "text"

    @property
    def topic_callbacks(self) -> dict[str, TopicCallback]:
        """Return a dictionary of topic callbacks."""
        result = super().topic_callbacks
        if self.command_topic and self.on_command:
            result[self.command_topic] = self.on_command
        return result


@attrs.define()
class MQTTSensorEntity(MQTTEntity):
    """A Home Assistant Sensor entity."""

    force_update: bool = False
    """Sends update events even if the value hasn't changed. Useful if you want to have
    meaningful value graphs in history."""
    suggested_display_precision: int = 0
    """The number of decimals which should be used in the sensor's state after rounding."""

    platform = "sensor"


@attrs.define()
class MQTTBinarySensorEntity(MQTTEntity):
    """A Home Assistant Binary Sensor entity."""

    payload_on: str = BOOL_ON
    payload_off: str = BOOL_OFF

    platform = "binary_sensor"


@attrs.define()
class MQTTSelectEntity(MQTTRWEntity):
    """A HomeAssistant Select entity."""

    options: Sequence[str] = attrs.field(default=None, validator=required)

    platform = "select"


@attrs.define()
class MQTTSwitchEntity(MQTTRWEntity):
    """A Home Assistant Switch entity."""

    payload_on: str = BOOL_ON
    payload_off: str = BOOL_OFF

    platform = "switch"


@attrs.define()
class MQTTTextEntity(MQTTRWEntity):
    """A Home Assistant Switch entity."""

    platform = "text"


@attrs.define()
class MQTTLightEntity(MQTTRWEntity):
    """A Home Assistant Switch entity."""

    payload_on: str = BOOL_ON
    payload_off: str = BOOL_OFF

    brightness_state_topic: str = ""
    brightness_command_topic: str = ""
    on_brightness_change: TopicCallback | None = None

    effect_state_topic: str = ""
    effect_command_topic: str = ""
    on_effect_change: TopicCallback | None = None
    effect_list: list[str] | None = None

    hs_state_topic: str = ""
    hs_command_topic: str = ""
    on_hs_change: TopicCallback | None = None

    platform = "light"

    async def send_brightness(
        self, client: MQTTAsyncClient, brightness: int, *, retain: bool = False
    ) -> None:
        """Publish the brightness to the MQTT brightness command topic."""
        await client.publish(
            self.brightness_state_topic,
            str(brightness),
            retain=retain,
        )

    async def send_effect(
        self, client: MQTTAsyncClient, effect: str, *, retain: bool = False
    ) -> None:
        """Publish the effect to the MQTT effect command topic."""
        await client.publish(
            self.effect_state_topic,
            effect,
            retain=retain,
        )

    async def send_hs(
        self, client: MQTTAsyncClient, hs: tuple[float, float], *, retain: bool = False
    ) -> None:
        """Publish the hue and saturation to the MQTT hs command topic."""
        await client.publish(
            self.hs_state_topic,
            f"{hs[0]},{hs[1]}",
            retain=retain,
        )

    @property
    def topic_callbacks(self) -> dict[str, TopicCallback]:
        """Return a dictionary of topic callbacks."""
        result = super().topic_callbacks
        if self.brightness_command_topic and self.on_brightness_change:
            result[self.brightness_command_topic] = self.on_brightness_change
        if self.effect_command_topic and self.on_effect_change:
            result[self.effect_command_topic] = self.on_effect_change
        if self.hs_command_topic and self.on_hs_change:
            result[self.hs_command_topic] = self.on_hs_change
        return result


@attrs.define()
class MQTTNumberEntity(MQTTRWEntity):
    """A HomeAssistant Number entity."""

    min: float = 0.0
    max: float = 100.0
    mode: str = "auto"
    step: float = 1.0

    suggested_display_precision: int = 0
    """The number of decimals which should be used in the sensor's state after rounding."""

    platform = "number"
